#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#define TRUE 1
#define FALSE 0

typedef struct object object;

typedef void (*free_proc)(object* o);

struct object {
  free_proc free;
};

void free_object(object *o) {
  o->free(o);
}

/* ------------------------------------ */

typedef struct dposn {
  object o;
  double x;
  double y;
} dposn;

void free_dposn(object* o) {
  free(o);
}

dposn* make_dposn(double x, double y) {
  dposn* dp;
  dp = (dposn*)malloc(sizeof(dposn));
  dp->o.free = free_dposn;
  dp->x = x;
  dp->y = y;
  return dp;
}

/* ------------------------------------ */

typedef struct shape shape;

typedef int (*pt_in_shape_proc)(dposn *dp, shape* s);

struct shape {
  object o;
  pt_in_shape_proc pt_in;
};

int pt_in_shape(dposn *dp, shape* s) {
  return s->pt_in(dp, s);
}

/* ------------------------------------ */

typedef struct circle {
  shape s;
  double radius;
  dposn* center;
} circle;

void free_circle(object* o) {
  circle *c = (circle *)o;
  free_object((object*)c->center);
  free(c);
}

double sqr(double v) { return v * v; }

int pt_in_circle(dposn *dp, shape* s) {
  circle* c = (circle*)s;
  return (sqrt(sqr(c->center->x - dp->x)
               + sqr(c->center->y - dp->y))
          <= c->radius);          
}

circle* make_circle(double radius, dposn* center) {
  circle* c;
  c = (circle*)malloc(sizeof(circle));
  c->s.o.free = free_circle;
  c->s.pt_in = pt_in_circle;
  c->radius = radius;
  c->center = make_dposn(center->x, center->y);
  return c;
}

/* ------------------------------------ */

typedef struct rectangle {
  shape s;
  dposn* topleft;
  double width;
  double height;
} rectangle;

int pt_in_rectangle(dposn* dp, shape* s) {
  rectangle* r = (rectangle*)s;
  return ((r->topleft->x <= dp->x)
          && (r->topleft->x + r->width >= dp->x)
          && (r->topleft->y <= dp->y)
          && (r->topleft->y + r->height >= dp->y));
}

void free_rectangle(object* o) {
  rectangle* r = (rectangle*)o;
  free_object((object*)r->topleft);
  free(r);
}

rectangle* make_rectangle(dposn* topleft, double width, double height) {
  rectangle* r;
  r = (rectangle*)malloc(sizeof(rectangle));
  r->s.o.free = free_rectangle;
  r->s.pt_in = pt_in_rectangle;
  r->topleft = make_dposn(topleft->x, topleft->y);
  r->width = width;
  r->height = height;
  return r;
}

/* ------------------------------------ */

void failed(int a, int b) {
  printf("test failed; expected %d, got %d\n", b, a);
  exit(1);
}

void check_int(int a, int b) {
  if (a != b) failed(a, b);
}

int main(int argc, char** argv)
{
  dposn* origin = make_dposn(0, 0);
  shape *c1 = (shape*)make_circle(3.5, origin);
  shape *r1 = (shape*)make_rectangle(origin, 3.2, 6.7);
  dposn* oneone = make_dposn(1, 1);
  dposn* onenone = make_dposn(1, -1);
  dposn* onefive = make_dposn(1, 5);

  check_int(pt_in_shape(oneone, c1), TRUE);
  check_int(pt_in_shape(oneone, r1), TRUE);

  check_int(pt_in_shape(onenone, c1), TRUE);
  check_int(pt_in_shape(onenone, r1), FALSE);

  check_int(pt_in_shape(onefive, c1), FALSE);
  check_int(pt_in_shape(onefive, r1), TRUE);

  free_object((object*)c1);
  free_object((object*)r1);
  free_object((object*)origin);
  free_object((object*)oneone);
  free_object((object*)onenone);
  free_object((object*)onefive);

  printf("ok\n");

  return 0;
}
